#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h"
#include"BMPIO.h"
#define PI 3.141592653589793238462643383279502884197

/* 正規化用の関数 */
void Normalize(Image *R,Image *G,Image *B){
  double max,min;
  max= min = R->img[0][0];
  int i,j;
  
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      if(max<R->img[i][j])max=R->img[i][j];
      if(max<G->img[i][j])max=G->img[i][j];
      if(max<B->img[i][j])max=B->img[i][j];
      if(min>R->img[i][j])min=R->img[i][j];
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];
    }
  double range = fabs(max-min);if(range<=0.0)range=1.0;
   
  for(i=0;i<R->sy;i++)
     for(j=0;j<R->sx;j++){
       double val = 255.0*(R->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      R->img[i][j] = val;
      val = 255.0*(G->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      G->img[i][j] = val;
      val = 255.0*(B->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      B->img[i][j] = val;
     }
     
}
/* 量子化 */
void Quantization(Image *in,Image *out,int binN){
  int i,j;
  double max=in->img[0][0];
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      if(max<in->img[i][j])max=in->img[i][j];
    }
  if(max<=0.0)max=1.0;
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      double val = ((double)(binN))*in->img[i][j]/max;
      int vali = ((int)(val));
      if(fabs(((double)(vali))-val)>=0.5)vali++;
      out->img[i][j] = ((double)(vali))*max/((double)(binN));
    }
}

/* maxで割る正規化 */
void Normalize(Image *in){
  int i,j;
  double max;
  max=in->img[0][0];
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++){
      if(max<in->img[i][j])max=in->img[i][j];
    }
  max=fabs(max);
  if(max<=0.0)max=1.0;
   for(i=0;i<in->sy;i++)
     for(j=0;j<in->sx;j++)in->img[i][j]/=max;
   
}


/* HSV量子化: Hue色相だけbinNの数に量子化、alphaは明度Vの強調パラメータ */
void QuantizationHSV(Image *R,Image *G,Image *B,int binN,double alpha){
  int i,j;
  
  Normalize(R,G,B);
  Image *H = new Image(R->sx,R->sy);
  Image *S = new Image(R->sx,R->sy);
  Image *V = new Image(R->sx,R->sy);
  
  
  //RGB->HSV
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      double max,min;
      int checkmax=0;
      min = max = R->img[i][j];
      if(max<G->img[i][j]){
	max=G->img[i][j];
	checkmax=1;
      }
      if(max<B->img[i][j]){
	max=B->img[i][j];
	checkmax=2;
      }
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];

      S->img[i][j] = (max-min);
      if(max>0.0)S->img[i][j]/=max;
      V->img[i][j] = max;
     
      double div=max-min;
      if(div<=0.0)div=1.0;
      if(checkmax==0){
	H->img[i][j] = 60.0*(G->img[i][j]-B->img[i][j])/div;
      }else if(checkmax==1){
	H->img[i][j] = 60.0*(B->img[i][j]-R->img[i][j])/div+120.0;
      }else{
	H->img[i][j] = 60.0*(R->img[i][j]-G->img[i][j])/div+240.0;
      }
      if(H->img[i][j]<0.0)H->img[i][j]+=360.0;
     
    }
  Normalize(S);
  Normalize(V);
  /* Vの強調 */ 
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      V->img[i][j] =  V->img[i][j]*20.0-10.0;
      V->img[i][j] = 255.0/(1.0+exp(-alpha*V->img[i][j]));
    }
  /* Hの量子化 */
  Quantization(H,H,binN);
  //HSV->RGB
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){

      if(S->img[i][j]==0.0){
	R->img[i][j]=G->img[i][j]=B->img[i][j]=V->img[i][j];
      }else{
      double Hi = H->img[i][j]/60.0;
      int Hii = ((int)(Hi));
      Hii = Hii % 6;
      
      double f = H->img[i][j]/60.0-((double)(Hii));
      double p = V->img[i][j]*(1.0-S->img[i][j]);
      double q = V->img[i][j]*(1.0-f*S->img[i][j]);
      double t = V->img[i][j]*(1.0-(1.0-f)*S->img[i][j]);
      if(Hii==0){
	R->img[i][j]=V->img[i][j];G->img[i][j]=t;B->img[i][j]=p;
      }else if(Hii==1){
	R->img[i][j]=q;G->img[i][j]=V->img[i][j];B->img[i][j]=p;
      }else if(Hii==2){
	R->img[i][j]=p;G->img[i][j]=V->img[i][j];B->img[i][j]=t;
      }else if(Hii==3){
	R->img[i][j]=p;G->img[i][j]=q;B->img[i][j]=V->img[i][j];
      }else if(Hii==4){
	R->img[i][j]=t;G->img[i][j]=p;B->img[i][j]=V->img[i][j];
      }else{
	R->img[i][j]=V->img[i][j];G->img[i][j]=p;B->img[i][j]=q;
      }
      }
      if(R->img[i][j]<0.0)R->img[i][j]=0.0;
      if(G->img[i][j]<0.0)G->img[i][j]=0.0;
      if(B->img[i][j]<0.0)B->img[i][j]=0.0;
      
      if(R->img[i][j]>255.0)R->img[i][j]=255.0;
      if(G->img[i][j]>255.0)G->img[i][j]=255.0;
      if(B->img[i][j]>255.0)B->img[i][j]=255.0;
      
    }
  delete H;
  delete S;
  delete V;

}



/* g++ -o ImageQuantization ImageQuantization.cxx -lm
引数5個
./ImageQuantization 入力ファイル名 出力ファイル名 HSV量子化数(int) HSV量子化V強調パラメータ(double) RGB量子化数(int) 

例えば、
./ImageQuantization lena.bmp style_lena.bmp 16 0.5 4

*/

int main(int argc,char *argv[]){
  int i,j;
  int sx,sy;
  /* BMP画像の読み込み */
  BMPIO *myio = new BMPIO();
  myio->readBMPSize(&sx,&sy,argv[1]);
  Image *R = new Image(sx,sy);
  Image *G = new Image(sx,sy);
  Image *B = new Image(sx,sy);
  myio->readBMP(R,G,B,argv[1]);
  

  /* 計算の為の画像 */
  
  Image *outR = new Image(sx,sy);
  Image *outG = new Image(sx,sy);
  Image *outB = new Image(sx,sy);
  
 
  /* HSVのHを多値化する数 */
  int HSV_N = atoi(argv[3]);
  /* HSV量子化のV強調パラメータ */
  double alpha = atof(argv[4]);
  /* RGBを多値化する数 */
  int RGB_N = atoi(argv[5]);

  
  
  /* HSV空間での多値化(量子化) */
  
  QuantizationHSV(R,G,B,HSV_N,alpha);
  
  /* RGB空間での多値化(量子化)　*/
  
  Quantization(R,outR,RGB_N);
  Quantization(G,outG,RGB_N);
  Quantization(B,outB,RGB_N);
  
 
  
  /* 出力量子化画像の保存 */
  myio->saveBMP(outR,outG,outB,argv[2]);
  
  /* メモリの開放 */
  
  delete R;
  delete G;
  delete B;

  delete outR;
  delete outG;
  delete outB;

  delete myio;

  return 0;
}

#undef PI
